# app/models/concerns/file_relocation_concern.rb
require 'json'
module S3FileHandleable
  extend ActiveSupport::Concern

  def relocate_files_on_folder_name_change
    return true unless saved_change_to_s3_folder_name?
    name_changes = saved_changes[:s3_folder_name]
    return true if name_changes[0].nil?

    Hyrax::HandleS3FolderNameJob.perform_later(self.class.name, self.id, name_changes)
  end

  def save_meta_json_in_s3
    s3 = S3StorageService.new
    s3.init_client

    crc_dataset = self.class.name == "CrcDataset" ? self.class.find(self.id) : self.crc_dataset
    bucket_name = s3.sanitise_name(crc_dataset.id)
    complex_modalities = self.class.name == "ComplexModality" ? [self] : self.complex_modalities.to_a.flatten

    complex_modalities.each do |modality|
      subject = modality.complex_subject
      session = modality.complex_session
      meta_file_key = "#{subject.s3_folder_name}/#{session.s3_folder_name}/#{modality.s3_folder_name}/meta.json"
      json_data = prepare_json(crc_dataset, subject, session, modality)

      s3.add_content(bucket_name, meta_file_key, json_data)
    end
  end

  private

  def prepare_json(crc_dataset, subject, session, modality)
    @meta = {}
    data_from_crc_dataset(crc_dataset)
    data_from_complex_subject(subject)
    data_from_complex_session(session)
    data_from_complex_modality(modality)
    JSON.pretty_generate(@meta)
  end

  def data_from_complex_subject(subject)
    @meta["Subject ID"] = ""
    @meta["Subject ID"] = subject.subject_title.strip if subject.subject_title.present?

    @meta["Subject Species"] = ""
    @meta["Subject Species"] = subject.subject_species.strip if subject.subject_species.present?

    @meta["Subject type"] = ""
    @meta["Subject type"] = subject.subject_type.strip if subject.subject_type.present?

    @meta["Subject sex"] = ""
    @meta["Subject sex"] = subject.subject_sex.strip if subject.subject_sex.present?

    @meta["Subject age"] = ""
    @meta["Subject age"] = subject.subject_age.to_s.strip if subject.subject_age.present?
  end

  def data_from_complex_session(session)
    # date from complex session will overwrite date from complex date
    if session.present? and session.session_date_recorded.present?
      @meta["Record date"] =  session.session_date_recorded
      @meta["DataCite-Date"] = session.session_date_recorded
      @meta["DataCite-dateType"] = "Created"
      @meta["DublinCore-Date"] = session.session_date_recorded
    end
  end

  def data_from_complex_modality(modality)
    @meta["Modality"] = []
    @meta["Modality"] << modality.modality if modality.present? and modality.modality.present?
    @meta["DataCite-ResourceType"] = @meta["Modality"]
  end

  def data_from_crc_dataset(crc_dataset)
    @meta["DublinCore-Title"] = crc_dataset.title[0]
    @meta["Experiment title"] = crc_dataset.title[0]
    @meta["DataCite-Title"] = crc_dataset.title[0]
    @meta["DataCite-titleType"] = ""
    @meta["DataCite-resourceTypeGeneral"] = "Collection"
    @meta["DublinCore-Type"] = "Collection"
    @meta["DublinCore-Source"] = ""
    

    # Language
    @meta["DublinCore-Language"] = ""
    @meta["DataCite-Language"] = ""
    if crc_dataset.language.any? and crc_dataset.language.first.present?
      @meta["DublinCore-Language"] = crc_dataset.language[0].strip
      @meta["DataCite-Language"] = crc_dataset.language[0].strip
    end

    # Size and Format
    @meta["DataCite-Size"] = crc_dataset.files_size.join("; ")
    @meta["DataCite-Format"] = crc_dataset.format.join("; ")
    @meta["DublinCore-Format"] = crc_dataset.format.join("; ")

    if crc_dataset.crc_resource_type.present?
      @meta["Resource Type"] = crc_dataset.crc_resource_type.strip
    end

    @meta["DataCite-Description"] = ""
    @meta["DublinCore-Description"] = ""
    @meta["DataCite-descriptionType"] = "Abstract"
    if crc_dataset.experiment_description.any? and crc_dataset.experiment_description.first.present?
      experiment_description = crc_dataset.experiment_description[0].strip
      @meta["DataCite-Description"] = experiment_description
      @meta["DublinCore-Description"] = experiment_description
      @meta["Experiment Description"] = experiment_description
    end

    @meta["Group ID"] = ""
    if crc_dataset.parent_collections.any? and
       crc_dataset.parent_collections.first.present? and
       crc_dataset.parent_collections.first.title.present?
      @meta["Group ID"] = crc_dataset.parent_collections.first.title[0].strip
    end
    @meta["Shared With"] = ""

    @meta["Animal|Ethics approval No."] = ""
    if crc_dataset.approval_number.present?
      @meta["Animal|Ethics approval No."] = crc_dataset.approval_number.to_s.strip
    end

    # Keywords/Subject
    @meta["DataCite-subjectSchemeURI"] = ""
    @meta["DataCite-subjectScheme"] = ""
    @meta["DataCite-subjectValueURI"] = ""
    @meta["DublinCore-Subject"] = ""
    @meta["DataCite-Subject"] = ""

    if crc_dataset.keyword.any?
      keywords = crc_dataset.keyword.reject(&:empty?).map { |s| s.strip }
      @meta["DublinCore-Subject"] = keywords.join("; ") if keywords.any?
      @meta["DataCite-Subject"] = keywords.join("; ") if keywords.any?
    end

    #Publisher
    @meta["DublinCore-Publisher"] = ""
    @meta["DataCite-Publisher"] = ""
    if crc_dataset.publisher.any? and crc_dataset.publisher.first.present?
      @meta["DublinCore-Publisher"] =  crc_dataset.publisher[0].strip
      @meta["DataCite-Publisher"] =  crc_dataset.publisher[0].strip
    end

    @meta["DublinCore-Coverage"] = ""
    if crc_dataset.coverage.present?
      coverage = crc_dataset.coverage.strip
      @meta["DublinCore-Coverage"] = coverage
      @meta["DataCite-GeoLocationPlace"] = coverage
    end

    @meta["Software version"] = crc_dataset.software_version.any? ? crc_dataset.software_version.first : ""
    @meta["DataCite-Version"] = "4.4"
    @meta["Extra information"] = ""
    if crc_dataset.extra_information.any? and crc_dataset.extra_information.first.present?
      @meta["Extra information"] = crc_dataset.extra_information[0].strip
    end

    # Rights
    @meta["DataCite-Rights"] = ""
    @meta["DublinCore-Rights"] = ""
    @meta["DataCite-rightsURI"] = ""
    @meta["DataCite-rightsIdentifierScheme"] = ""
    @meta["DataCite-rightsSchemeURI"] = ""
    @meta["DataCite-rightsIdentifier"] = ""

    if crc_dataset.license.any? and crc_dataset.license.first.present?
      rights = LicenseService.new.label(crc_dataset.license.first)
      @meta["DataCite-Rights"] = rights
      @meta["DublinCore-Rights"] = rights
      @meta["DataCite-rightsURI"] = crc_dataset.license.first
    end

    # Map ARK and DOI
    @meta["DataCite-alternateIdentifier"] = ""
    @meta["DataCite-alternateIdentifierType"] = ""
    @meta["DataCite-Identifier"] = ""
    @meta["DublinCore-Identifier"] = ""
    @meta["DataCite-identifierType"] = ""
    ark = ''
    
    crc_dataset.complex_identifier.each do |idf|
      if idf.scheme.any? and idf.scheme[0] == "ARK" and
         idf.identifier.any? and idf.identifier[0].present?
        ark = idf.identifier[0].strip
        @meta["DataCite-alternateIdentifier"] = ark
        @meta["DataCite-alternateIdentifierType"] = "ARK"
      elsif idf.scheme.any? and idf.scheme[0] == "DOI" and
            idf.identifier.any? and idf.identifier[0].present?
        @meta["DublinCore-Identifier"] = idf.identifier[0].strip
        @meta["DataCite-Identifier"] = idf.identifier[0].strip
        @meta["DataCite-identifierType"] = "DOI"
      end
    end
      
    if ark.empty?
      @meta["DataCite-alternateIdentifier"] = crc_dataset.id
      @meta["DataCite-alternateIdentifierType"] = "Local"
    end
    
    data_from_complex_person(crc_dataset.complex_person)
    data_from_complex_date(crc_dataset.complex_date)
    data_from_complex_funding_reference(crc_dataset.complex_funding_reference)
    data_from_complex_relation(crc_dataset.complex_relation)
  end

  def data_from_complex_person(complex_person)
    complex_creators, complex_contributors = group_people_by_role(complex_person)
    # Creators
    creators = get_people_info(complex_creators)
    # names, first_names, last_names, affiliations, name_identifiers, roles
    @meta["Creator"] = creators[:names]
    @meta["DublinCore-Creator"] = creators[:names]
    @meta["DataCite-creatorName"] = creators[:names]
    @meta["DataCite-creatorNameType"] = creators[:name_types]
    @meta["DataCite-creatorFamilyName"] = creators[:last_names]
    @meta["DataCite-creatorGivenName"] = creators[:first_names]
    @meta["DataCite-creatorAffiliation"] = creators[:affiliations]
    @meta["DataCite-creatorNameIdentifier"] = creators[:name_identifiers]
    @meta["DataCite-creatorNameIdentifierScheme"] = creators[:name_identifier_schemes]
    @meta["DataCite-creatorSchemeURI"] = creators[:scheme_uri]

    # Contributors
    contributors = get_people_info(complex_contributors)
    # names, first_names, last_names, affiliations, name_identifiers, roles
    @meta["Contributor"] = contributors[:names]
    @meta["DublinCore-Contributor"] = contributors[:names]
    @meta["DataCite-contributorName"] = contributors[:names]
    @meta["DataCite-contributorNameType"] = contributors[:name_types]
    @meta["DataCite-contributorFamilyName"] = contributors[:last_names]
    @meta["DataCite-contributorGivenName"] = contributors[:first_names]
    @meta["DataCite-contirbutorAffiliation"] = contributors[:affiliations]
    @meta["DataCite-contributorType"] = contributors[:roles]
    @meta["DataCite-contributorNameIdentifier"] = contributors[:name_identifiers]
    @meta["DataCite-contributorNameIdentifierScheme"] = contributors[:name_identifier_schemes]
    @meta["DataCite-contributorSchemeURI"] = contributors[:scheme_uri]
  end

  def group_people_by_role(complex_person)
    creators = []
    contributors = []
    complex_person.each do |person|
      role = person.role[0].strip if person.role.any? and person.role.first.present?
      if role == "creator"
        creators << person
      else
        contributors << person
      end
    end
    return creators, contributors
  end

  def get_people_info(complex_people)
    names = []
    first_names = []
    last_names = []
    affiliations = []
    name_identifiers = []
    roles = []
    name_types = []
    name_identifier_schemes = []
    scheme_uri = []

    complex_people.each do |person|
      full_name = []
      if person.last_name.any? and person.last_name.first.present?
        full_name << person.last_name[0].strip
        last_names << person.last_name[0].strip
      else
        full_name << ""
        last_names << ""
      end
      if person.first_name.any? and person.first_name.first.present?
        full_name << person.first_name[0].strip
        first_names << person.first_name[0].strip
      else
        first_names << ""
      end

      names << full_name.join(', ') if full_name.any?
      name_types << "Personal"
      if person.affiliation.any? and person.affiliation.first.present?
        affiliations << person.affiliation[0].strip
      else
        affiliations << ""
      end

      if person.orcid.any? and person.orcid.first.present?
        name_identifiers << person.orcid[0].strip
        name_identifier_schemes << "Orcid"
        scheme_uri << "https://orcid.org"
      else
        name_identifiers << ""
        name_identifier_schemes << ""
        scheme_uri << ""
      end

      if person.role.any? and person.role.first.present?
        roles << person.role[0].strip
      else
        roles << ""
      end
    end

    people = {
      names: names.join("; "),
      first_names: first_names.join("; "),
      last_names: last_names.join("; "),
      affiliations: affiliations.join("; "),
      name_identifiers: name_identifiers.join("; "),
      roles: roles.join("; "),
      name_types: name_types.join("; "),
      name_identifier_schemes: name_identifier_schemes.join("; "),
      scheme_uri: scheme_uri.join("; ")
    }
    people
  end

  def data_from_complex_funding_reference(complex_funding_reference)
    funder_name = []
    funder_identifier = []

    complex_funding_reference.each do |funder|
      if funder.funder_identifier.any? and funder.funder_identifier.first.present?
        funder_identifier << funder.funder_identifier[0].strip
      end

      if funder.funder_name.any? and funder.funder_name.first.present?
        funder_name << funder.funder_name[0].strip
      end
    end

    @meta["DataCite-funderIdentifier"] = funder_identifier.join("; ")
    @meta["DataCite-funderName"] = funder_name.join("; ")
    @meta["DataCite-FundingReference"] = "DFG"
    @meta["DataCite-funderIdentifierType"] = "Crossref Funder ID"
  end

  def data_from_complex_date(complex_dates)
    @meta["Record date"] = ""
    @meta["DataCite-Date"] = ""
    @meta["DataCite-dateType"] = ""
    @meta["DublinCore-Date"] = ""
    @meta["DataCite-PublicationYear"] = ""
    date = []
    record_date = []
    date_type = []

    complex_dates.each do |complex_date|
      if complex_date.description.any? and complex_date.description.first.present? and
        complex_date.date.any? and complex_date.date.first.present?
        term = DateService.new.find_by_id(complex_date.description.first)

        if term["label"] == "Recorded"
          record_date << complex_date.date[0]
        elsif term["label"] == "Created"
          date << complex_date.date[0]
          date_type << 'Created'
        elsif term["label"] == "Published" 
          @meta["DataCite-PublicationYear"] = complex_date.date[0].to_date.year.to_s
        end
      end
    end

    @meta["Record date"] = record_date.join(';')
    @meta["DataCite-Date"] = date.join(';')
    @meta["DataCite-dateType"] = date_type.join(';')
    @meta["DublinCore-Date"] = date.join(';')
  end

  def data_from_complex_relation(complex_relations)
    @meta["DublinCore-Relation"] = ""
    @meta["DataCite-RelatedIdentifier"] = ""
    @meta["DataCite-relatedIdentifierType"] = ""
    @meta["DataCite-relationType"] = ""
    urls = []
    relations = []

    complex_relations.each do |complex_relation|
      urls << complex_relation.url[0].strip if complex_relation.url.any?
      relations << complex_relation.relationship[0].strip if complex_relation.relationship.any?
    end

    relation_urls = urls.join("; ")

    @meta["DublinCore-Relation"] = relation_urls
    @meta["DataCite-RelatedIdentifier"] = relation_urls
    @meta["DataCite-relationType"] = relations.join("; ") if relations.any?
    @meta["DataCite-relatedIdentifierType"] = relation_urls.downcase.include?('doi') ? "DOI" : "URL"
  end

end
